﻿//-----------------------------------------------------------------------
// InfoPath Form Extractor - "IFE.ashx.cs" - by SpittingCAML (blog.mgallen.com)
// No warranty provided, or implied. Use at own risk.
// This code is not production ready, and is provided for discussion only.
// Acknoledgements given in each method definition.
//-----------------------------------------------------------------------
namespace SC.Demo.SharePoint.Services
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.IO;
    using System.Security.Cryptography;
    using System.Security.Principal;
    using System.Text;
    using System.Web;
    using System.Web.Services;
    using System.Xml;
    using Microsoft.SharePoint;
    using Microsoft.SharePoint.Administration;
    using Microsoft.SharePoint.Utilities;

    /// <summary>
    /// Summary description for InfoPathFileExtractor (IFE)
    /// </summary>
    [WebService(Namespace = "http://sc.demo.sharepoint.services.ife/")]
    [WebServiceBinding(ConformsTo = WsiProfiles.BasicProfile1_1)]
    public class IFE : IHttpHandler
    {
        #region member variables <set up on initialize>
        /// <summary>
        /// this is the name of the field group storing required information
        /// </summary>
        private string _ifeGroupName;

        /// <summary>
        /// this is the name of the field group storing the user form data
        /// </summary>
        private string _rootGroupName;

        /// <summary>
        /// this is the infopath form retrieved from the http POST in XmlDocument form
        /// </summary>
        private XmlDocument _ipForm;

        /// <summary>
        /// this is the infopath form as a string
        /// </summary>
        private string _ipFormString;

        /// <summary>
        /// this is the InfoPath filename (that supplied InfoPath form will be saved as) as string
        /// </summary>
        private string _ipSaveFileName;

        /// <summary>
        /// this is the SharePoint site url as a string
        /// </summary>
        private string _spSiteURL;

        /// <summary>
        /// this is the Form Library url as a string
        /// </summary>
        private string _spFormLibraryURL;

        /// <summary>
        /// this is the Document Library url as a string
        /// </summary>
        private string _spDocumentLibraryURL;

        /// <summary>
        /// this is the folder to create the file in as a string
        /// </summary>
        private string _documentLibraryFolder;

        /// <summary>
        /// this is the output from the webhandler
        /// </summary>
        private string _outputConsole;

        /// <summary>
        /// this is a boolean indicating whether files should be overwritten
        /// </summary>
        private bool _overwriteFilesIfExist;

        /// <summary>
        /// this is the identity of the user who instigated this handler
        /// </summary>
        private WindowsIdentity _userIdentity;
        #endregion

        /// <summary>
        /// this method is invoked when the handler is requested
        /// 
        /// N.B: the maxRequestLength set as part of the application pool
        /// must be set large enough to receive files > 4MB in size as
        /// the default value is very small.
        /// </summary>
        /// <param name="context"></param>
        public void ProcessRequest(HttpContext context)
        {
            string strmContents = string.Empty;

            // extract the XmlData from the inputstream
            using (System.IO.StreamReader reader = new System.IO.StreamReader(context.Request.InputStream))
            {
                while (reader.Peek() >= 0)
                {
                    strmContents += reader.ReadLine();
                }
            }

            // initialise this class
            this.Initialize(context, strmContents);

            if (this.Initialized())
            {
                // process infopath form
                this.StartProcess(this._ipForm);
            }
            else
            {
                this._outputConsole += "Initialization failed<br/>";
            }

            // provide response to user (can view response in InfoPath)
            context.Response.ContentType = "text/plain";
            context.Response.Write(this._outputConsole);
        }

        /// <summary>
        /// this determines the reusable state of this handler
        /// </summary>
        public bool IsReusable
        {
            get
            {
                return false;
            }
        }

        #region private static methods
        /// <summary>
        /// this method returns all namespaces in scopt
        /// </summary>
        /// <param name="xmlDocument">the inputted document</param>
        /// <returns>an XmlNamespaceManager object containing namespaces in scope</returns>
        private static XmlNamespaceManager GetDocumentElementNamespaces(XmlDocument xmlDocument)
        {
            if (xmlDocument == null)
            {
                HandleException(new ArgumentNullException("xmlDocument"));
            }

            XmlNamespaceManager manager = new XmlNamespaceManager(xmlDocument.NameTable);
            if (xmlDocument.DocumentElement != null)
            {
                IDictionary<string, string> namespacesInScope = xmlDocument.DocumentElement.CreateNavigator().GetNamespacesInScope(XmlNamespaceScope.All);
                if (namespacesInScope == null)
                {
                    return manager;
                }

                foreach (string str in namespacesInScope.Keys)
                {
                    if (!manager.HasNamespace(str))
                    {
                        manager.AddNamespace(str, namespacesInScope[str]);
                    }
                }
            }

            return manager;
        }

        /// <summary>
        /// this method retrieves a specific node from an xmldocument/infopath form XML
        /// </summary>
        /// <param name="ipForm">an xml document or infopath form</param>
        /// <param name="nodeName">the name of the node to retrieve</param>
        /// <returns>an XmlNode object for the requested node, or null if not found</returns>
        private static XmlNode GetSpecificNode(XmlDocument ipForm, string nodeName)
        {
            if (ipForm != null)
            {
                Dictionary<string, string> namespacesInScope = GetDocumentElementNamespaces(ipForm).GetNamespacesInScope(XmlNamespaceScope.All) as Dictionary<string, string>;
                foreach (string str in namespacesInScope.Values)
                {
                    if (ipForm.GetElementsByTagName(nodeName, str).Count > 0)
                    {
                        return ipForm.GetElementsByTagName(nodeName, str).Item(0);
                    }
                }
            }

            return null;
        }

        /// <summary>
        /// this method retrieves the value of a specified node and field name
        /// </summary>
        /// <param name="specifiedNode">a node object</param>
        /// <param name="fieldName">the requested field</param>
        /// <returns>string containing the value, or string.Empty if not found</returns>
        private static string GetSpecifiedNodeFieldValue(XmlNode specifiedNode, string fieldName)
        {
            if ((specifiedNode == null) || string.IsNullOrEmpty(fieldName))
            {
                return string.Empty;
            }

            XmlNamespaceManager nsmgr = new XmlNamespaceManager(specifiedNode.OwnerDocument.NameTable);
            nsmgr.AddNamespace(specifiedNode.Prefix, specifiedNode.NamespaceURI);

            // concatenate prefix onto fieldname
            string xpath = specifiedNode.Prefix + ":" + fieldName;

            if (specifiedNode.SelectSingleNode(xpath, nsmgr) == null)
            {
                return string.Empty;
            }

            return specifiedNode.SelectSingleNode(xpath, nsmgr).InnerText.Trim();
        }

        /// <summary>
        /// this method handles exceptions
        /// </summary>
        /// <param name="ex">an exeption object</param>
        private static void HandleException(Exception ex)
        {
            throw ex;
        }

        /// <summary>
        /// method for validating a url with regular expressions
        /// </summary>
        /// <param name="url">url we're validating</param>
        /// <returns>true if valid, otherwise false</returns>
        private static bool IsValidUrl(string url)
        {
            // commented out as this does not work with localhost and non fqdn addresses
            //string pattern = @"^(http|https|ftp)\://[a-zA-Z0-9\-\.]+\.[a-zA-Z]{2,3}(:[a-zA-Z0-9]*)?/?([a-zA-Z0-9\-\._\?\,\'/\\\+&amp;%\$#\=~])*[^\.\,\)\(\s]$";
            //System.Text.RegularExpressions.Regex reg = new System.Text.RegularExpressions.Regex(pattern, System.Text.RegularExpressions.RegexOptions.Compiled | System.Text.RegularExpressions.RegexOptions.IgnoreCase);
            //return reg.IsMatch(url);

            return System.Uri.IsWellFormedUriString(url, UriKind.Absolute);
        }

        /// <summary>
        /// this method ensures InfoPath files are saved with a valid filename
        /// </summary>
        /// <param name="ipFileName">the incoming infopath filename</param>
        /// <returns>true if valid, false if not valid</returns>
        private static bool IsValidInfoPathFilename(string ipFileName)
        {
            return ipFileName.EndsWith(".xml");
        }

        /// <summary>
        /// this method ensures each URL has a trailing slash
        /// </summary>
        /// <param name="url">the url to check</param>
        /// <returns>url with trailing slash</returns>
        private static string AddTrailingSlash(string url)
        {
            if (!url.EndsWith("/"))
            {
                url = url + "/";
            }

            return url;
        }
        #endregion

        /// <summary>
        /// this method reports whether this class has been successfully initialised
        /// </summary>
        /// <returns>true if it has, false if it hasn't</returns>
        private bool Initialized()
        {
            bool init = true;

            init = !string.IsNullOrEmpty(this._ipSaveFileName) && !string.IsNullOrEmpty(this._spDocumentLibraryURL)
                && !string.IsNullOrEmpty(this._spFormLibraryURL) && !string.IsNullOrEmpty(this._spSiteURL)
                && !string.IsNullOrEmpty(this._documentLibraryFolder);

            return init;
        }

        /// <summary>
        /// this method initialises this handler
        /// </summary>
        /// <param name="ipFormString">the infopath form as a string</param>
        private void Initialize(HttpContext context, string ipFormString)
        {
            // clear output console
            this._outputConsole = string.Empty;

            // get identity of current user
            this._userIdentity = (System.Security.Principal.WindowsIdentity)context.User.Identity;

            if (string.IsNullOrEmpty(ipFormString))
            {
                this._outputConsole += "No InfoPath form provided in request";
                HandleException(new ArgumentNullException("entireForm"));
            }
            else
            {
                // Example query string: http://myserver/myapp/IFE.ashx?RootGroup=myFields&IFEGroup=IFE
                //
                // Query String will determine 
                //          - IFE Group Name (a group of fields used to configure this webhandler) [KEY: IFEGroup]
                //          - Root Group Name (the root group of user supplied fields in InfoPath, by default this is 'myfields') [KEY: RootGroup]
                //                            typically this root group will contain all the file attachments
                //
                // Using these query string parameters the system is able to ascertain all information required to perform its job

                // set up query string related configuration
                this.RetrieveQueryStringItems(context);

                // Extracted from the RootGroup
                //          - All file attachments
                //
                // Extracted from IFEGroup
                //
                // SharePoint site URL - [InfoPath field name: SiteURL] { Valid URL }
                // InfoPath form library URL - [InfoPath field name: FormLibraryURL] { Valid URL }
                // Document Library URL - [InfoPath field name: DocumentLibraryURL] { Valid URL }
                // Document Library Folder - [InfoPath field name: DocumentLibraryFolder] { Valid URL }
                // Name that the InfoPath form will be stored as - [InfoPath field name: InfoPathSaveAsFileName] { Valid InfoPath filename }
                // Whether files should be overwritten - [InfoPath field name: OverwriteFiles] {'True' or 'False'}

                this._ipFormString = ipFormString;
                this._ipForm = new XmlDocument();
                this._ipForm.PreserveWhitespace = true;
                this._ipForm.LoadXml(this._ipFormString);
                XmlNode ifeNode = GetSpecificNode(this._ipForm, this._ifeGroupName);
                XmlNode rootNode = GetSpecificNode(this._ipForm, this._rootGroupName);

                if (IsValidUrl(GetSpecifiedNodeFieldValue(ifeNode, "SiteURL")))
                {
                    this._spSiteURL = AddTrailingSlash(GetSpecifiedNodeFieldValue(ifeNode, "SiteURL"));
                    this._outputConsole += "SharePoint Site: " + this._spSiteURL + "<br/>";
                }
                else
                {
                    this._outputConsole += "SharePoint Site: value is invalid.<br/>";
                }

                if (IsValidUrl(GetSpecifiedNodeFieldValue(ifeNode, "FormLibraryURL")))
                {
                    this._spFormLibraryURL = AddTrailingSlash(GetSpecifiedNodeFieldValue(ifeNode, "FormLibraryURL"));
                    this._outputConsole += "InfoPath Form Library: " + this._spFormLibraryURL + "<br/>";
                }
                else
                {
                    this._outputConsole += "InfoPath Form Library: value is invalid.<br/>";
                }

                if (IsValidUrl(GetSpecifiedNodeFieldValue(ifeNode, "DocumentLibraryURL")))
                {
                    this._spDocumentLibraryURL = AddTrailingSlash(GetSpecifiedNodeFieldValue(ifeNode, "DocumentLibraryURL"));
                    this._outputConsole += "Document Library: " + this._spDocumentLibraryURL + "<br/>";
                }
                else
                {
                    this._outputConsole += "Document Library: value is invalid.<br/>";
                }

                this._documentLibraryFolder = GetSpecifiedNodeFieldValue(ifeNode, "DocumentLibraryFolder");

                if (!string.IsNullOrEmpty(this._documentLibraryFolder))
                {
                    // add a trailing slash to the document library folder as it is used in concatenations
                    this._documentLibraryFolder = AddTrailingSlash(this._documentLibraryFolder);
                }

                this._outputConsole += "Document Library Folder: " + this._documentLibraryFolder + "<br/>";

                if (IsValidInfoPathFilename(GetSpecifiedNodeFieldValue(ifeNode, "InfoPathSaveAsFileName")))
                {
                    this._ipSaveFileName = GetSpecifiedNodeFieldValue(ifeNode, "InfoPathSaveAsFileName");
                    this._outputConsole += "InfoPath Save FileName: " + this._ipSaveFileName + "<br/>";
                }
                else
                {
                    this._outputConsole += "InfoPath Save FileName: value is invalid<br/>";
                }

                bool outBool;
                if (bool.TryParse(GetSpecifiedNodeFieldValue(ifeNode, "OverwriteFiles"), out outBool))
                {
                    this._overwriteFilesIfExist = bool.Parse(GetSpecifiedNodeFieldValue(ifeNode, "OverwriteFiles"));
                    if (this._overwriteFilesIfExist)
                    {
                        this._outputConsole += "WARNING: overwriting files that already exist" + "<br/>";
                    }
                    else
                    {
                        this._outputConsole += "Not overwriting files that already exist" + "<br/>";
                    }
                }
                else
                {
                    // default to false
                    this._overwriteFilesIfExist = false;
                    this._outputConsole += "Not overwriting files that already exist" + "<br/>";
                }
            }
        }

        /// <summary>
        /// this method scans the infopath form for base64 files
        /// and sends them to the file processing method
        /// </summary>
        /// <param name="ipForm">the input infopath form</param>
        private void StartProcess(XmlDocument ipForm)
        {
            // start at the root, as supplied by the query string
            XmlNode rootNode = GetSpecificNode(ipForm, this._rootGroupName);
            this.FindFileAttachments(rootNode.ChildNodes);

            // get the xml form as a string to encode
            StringWriter sw = new StringWriter();
            XmlTextWriter xw = new XmlTextWriter(sw);
            this._ipForm.WriteTo(xw);

            // encode InfoPath form
            UTF8Encoding uenc = new UTF8Encoding();
            byte[] encodedIPForm = uenc.GetBytes(sw.ToString());

            // upload infopath form to the form library
            this.UploadFile(this._spFormLibraryURL + this.CleanFileName(this._ipSaveFileName), encodedIPForm, null);
        }

        /// <summary>
        /// this is a recursive method that will find all base64 encoded fields
        /// </summary>
        /// <param name="nodes">a list of children from the root node</param>
        private void FindFileAttachments(XmlNodeList nodes)
        {
            foreach (XmlNode node in nodes)
            {
                if (node.NodeType == XmlNodeType.Element)
                {
                    string fieldName = node.Name;
                    string localName = node.LocalName;
                    string fieldValue = node.InnerText;

                    if (IsBase64Data(fieldValue))
                    {
                        this.AttachmentProcessor(fieldValue, fieldName, localName);
                    }
                }

                this.FindFileAttachments(node.ChildNodes);
            }
        }

        /// <summary>
        /// this method processes each infopath attachment
        /// </summary>
        /// <param name="fieldValue">the infopath file attachment</param>
        /// <param name="fieldName">the name of the Xml field [with namespace prefix]</param>
        /// <param name="localName">the name of the Xml field [without namespace prefix]</param>
        private void AttachmentProcessor(string fieldValue, string fieldName, string localName)
        {
            // decode the attachment
            InfoPathAttachmentDecoder ipad = new InfoPathAttachmentDecoder(fieldValue);
            string fileName = this.CleanFileName(ipad.Filename);
            byte[] binary = ipad.DecodedAttachment;

            this._outputConsole += "Reading File: " + fileName + "<br/>";

            // only extract files with a valid extension (as defined in the sharepoint administration pages)
            if (IsValidFileExtension(this._spSiteURL, ipad.Filename))
            {
                // ensure we don't replace the shortcuts on update to existing forms
                if (!fileName.ToLower().EndsWith("url"))
                {
                    // metadata setup
                    Hashtable ht = this.SetupMetadata(fileName);

                    // upload to sharepoint
                    this.UploadFile(this._spDocumentLibraryURL + this._documentLibraryFolder + fileName, binary, ht);

                    // create lnk/url file
                    byte[] lnkFileBytes = this.CreateUrlLink(this._spDocumentLibraryURL + this._documentLibraryFolder + fileName);

                    // encode lnk file
                    InfoPathAttachmentEncoder ipae = new InfoPathAttachmentEncoder(lnkFileBytes, ipad.Filename + ".url", lnkFileBytes.Length);
                    string encoded = ipae.ToBase64String();

                    // replace original attachment with lnk file
                    XmlNode originalAttachmentNode = GetSpecificNode(this._ipForm, localName);
                    originalAttachmentNode.InnerText = encoded;

                    // dispose of InfoPath Attachment Encoder
                    ipae = null;

                    this._outputConsole += "Processed File: " + fileName + "<br/>";
                }
                else
                {
                    this._outputConsole += "File is already in shortcut format - no need to process: " + fileName + "<br/>";
                }
            }
            else
            {
                this._outputConsole += "File extension type not allowed: " + fileName + "<br/>";
            }

            // dispose of InfoPath Attachment Decoder object
            ipad = null;
        }

        /// <summary>
        /// this method configures the metadata for files uploaded into SharePoint
        /// </summary>
        /// <param name="filename">the filename of the file</param>
        /// <returns>Hashtable containing metadata</returns>
        private Hashtable SetupMetadata(string filename)
        {
            Hashtable ht = new Hashtable();

            ht.Add("Title", filename);
            ht.Add("Author", this._userIdentity.Name);
            ht.Add("ContentType", "Document");

            return ht;
        }

        /// <summary>
        /// this method creates a link file and returns it as UTF8 encoded byte array
        /// </summary>
        /// <param name="documentLocation">the url that the document being replaced with this link is stored</param>
        /// <returns>byte array containing UTF8 encoded *.url file</returns>
        private byte[] CreateUrlLink(string documentLocation)
        {
            string shortcut = string.Empty;
            shortcut += "[InternetShortcut]" + Environment.NewLine;
            shortcut += "URL=" + documentLocation;

            Encoding enc = Encoding.UTF8;
            return enc.GetBytes(shortcut);
        }

        /// <summary>
        /// this method determines whether a string is base64 encoded file
        /// this is may not be the best way of doing this
        /// </summary>
        /// <param name="inputString">the string to validate</param>
        /// <returns>true, if base64, false if not base64</returns>
        private bool IsBase64Data(string inputString)
        {
            // if empty string - this can't be a file!
            if (string.IsNullOrEmpty(inputString))
            {
                return false;
            }

            if ((inputString.Length % 4) != 0)
            {
                return false;
            }

            // Check that the string matches the base64 layout
            System.Text.RegularExpressions.Regex regex = new System.Text.RegularExpressions.Regex(@"[^-A-Za-z0-9+/=]|=[^=]|={3,}$");
            if (regex.Match(inputString).Success)
            {
                return false;
            }

            try
            {
                // If no exception is caught during decode, then it is possibly a base64 encoded string from InfoPath
                InfoPathAttachmentDecoder ipad = new InfoPathAttachmentDecoder(inputString);

                return true;
            }
            catch
            {
                // If exception is caught, then it is not a base64 encoded string
                return false;
            }
        }

        /// <summary>
        /// this method retrieves the query strings required for this handler to function
        /// </summary>
        /// <param name="context">a HttpContext object</param>
        private void RetrieveQueryStringItems(HttpContext context)
        {
            if (context != null)
            {
                if (context.Request["IFEGroup"] != null)
                {
                    this._ifeGroupName = context.Request["IFEGroup"];
                }
                else
                {
                    this._ifeGroupName = string.Empty;
                }

                if (context.Request["RootGroup"] != null)
                {
                    this._rootGroupName = context.Request["RootGroup"];
                }
                else
                {
                    this._rootGroupName = string.Empty;
                }
            }
        }

        /// <summary>
        /// The EnsureParentFolder method ensures that the parent folder in the destination URL 
        /// exists in the specified site, and it returns the site-relative URL of the parent folder. 
        /// The EnsureParentFolder method accepts two parameters: an SPWeb object that represents the parent site,
        /// and a string that contains the absolute URL that is passed from the UploadFile method. 
        /// If the parent folder does not exist, the EnsureParentFolder method creates it
        /// </summary>
        /// <param name="parentWeb">the parent site (web)</param>
        /// <param name="destinUrl">the destination url</param>
        /// <returns>the parent folder url</returns>
        public string EnsureParentFolder(string destinUrl)
        {
            string parentFolderUrl = string.Empty;

            SPSecurity.RunWithElevatedPrivileges(delegate()
            {
                using (SPSite destSite = new SPSite(destinUrl))
                {
                    using (SPWeb parentWeb = destSite.OpenWeb())
                    {
                        parentWeb.AllowUnsafeUpdates = true;
                        destinUrl = parentWeb.GetFile(destinUrl).Url;
                        int index = destinUrl.LastIndexOf("/");

                        if (index > -1)
                        {
                            parentFolderUrl = destinUrl.Substring(0, index);
                            SPFolder parentFolder = parentWeb.GetFolder(parentFolderUrl);

                            if (!parentFolder.Exists)
                            {
                                SPFolder currentFolder = parentWeb.RootFolder;

                                foreach (string folder in parentFolderUrl.Split('/'))
                                {
                                    currentFolder = currentFolder.SubFolders.Add(folder);
                                }
                            }
                        }
                    }
                }
            });

            return parentFolderUrl;
        }

        /// <summary>
        /// this method checks that the document currently in infopath does not
        /// violate the rules on filetypes currently excluded from SharePoint e.g. *.exe
        /// 
        /// thanks to: http://www.myfatblog.co.uk/?p=102
        /// </summary>
        /// <param name="currentSite">the current site url</param>
        /// <param name="filename">the filename to check</param>
        /// <returns>true if file extension is ok, false if it is excluded by system admins</returns>
        public bool IsValidFileExtension(string currentSiteUrl, string filename)
        {
            bool returnValue = false;

            SPSecurity.RunWithElevatedPrivileges(delegate()
            {
                using (SPSite currentSite = new SPSite(currentSiteUrl))
                {
                    using (SPWeb currentWeb = currentSite.OpenWeb())
                    {
                        Uri siteURI = new Uri(currentSite.Url);
                        SPWebApplication webApp = SPWebApplication.Lookup(siteURI);

                        Collection<string> blockedTypes = webApp.BlockedFileExtensions;

                        string[] fileExt = filename.Split('.');
                        string strExtension = fileExt[fileExt.Length - 1];

                        if (blockedTypes.Contains(strExtension))
                        {
                            returnValue = false;
                        }
                        else
                        {
                            returnValue = true;
                        }
                    }
                }
            });

            return returnValue;
        }

        /// <summary>
        /// Validates a file name before upload to the sharepoint library.
        /// 
        /// thanks to: http://developers.de/blogs/adis_jugo/archive/2009/10/29/validate-a-file-name-before-upload-to-the-sharepoint-library.aspx
        /// </summary>
        /// <param name="fileName">Original file name</param>
        /// <returns>Clean file name</returns>
        private string CleanFileName(string fileName)
        {
            string ext = Path.GetExtension(fileName);
            string name = Path.GetFileNameWithoutExtension(fileName);

            //first remove double fullstops (..)
            while (name.Contains(".."))
            {
                name = name.Replace("..", "__");
            }

            //name only is not allowed to end with fullstop (.)
            if (name.EndsWith("."))
            {
                name = name.Substring(0, name.Length - 1) + "_";
            }

            //now use "SPEncode.IsLegalCharInUrl()" to remove characters which sharepoint does not like
            char[] filenameChars = name.ToCharArray();
            foreach (char c in filenameChars)
            {
                if (!SPEncode.IsLegalCharInUrl(c)) name = name.Replace(c.ToString(), "");
            }

            //return clean filename
            return string.Format("{0}{1}", name, ext);
        }

        /// <summary>
        /// this method uploads the file
        /// 
        /// N.B: The size of the file that is uploaded cannot exceed two gigabytes
        /// </summary>
        /// <param name="destUrl">the destination url (including filename) e.g. http://mysite/mylib/test.txt</param>
        /// <param name="fileContents">the file contents (as byte array)</param>
        /// <param name="metadata">a hashtable containing metadata for the file</param>
        private void UploadFile(string destUrl, byte[] fileContents, object metadata)
        {
            // make sure destination folder will exist
            this.EnsureParentFolder(destUrl);

            using (SPSite destSite = new SPSite(destUrl))
            {
                using (SPWeb currentWeb = destSite.OpenWeb())
                {
                    currentWeb.AllowUnsafeUpdates = true;
                 
                    if (metadata != null)
                    {
                        try
                        {
                            WindowsImpersonationContext context = null;
                            try
                            {
                                context = this._userIdentity.Impersonate();
                                currentWeb.Files.Add(destUrl, fileContents, (Hashtable)metadata, this._overwriteFilesIfExist);
                            }
                            catch (Exception exception)
                            {
                                HandleException(exception);
                            }
                            finally
                            {
                                if (context != null)
                                {
                                    context.Undo();
                                }
                            }
                        }
                        catch (Exception ex)
                        {
                            this._outputConsole += ex.Message + "<br/>";
                        }
                    }
                    else
                    {
                        try
                        {
                            WindowsImpersonationContext context = null;
                            try
                            {
                                context = this._userIdentity.Impersonate();
                                currentWeb.Files.Add(destUrl, fileContents, this._overwriteFilesIfExist);
                            }
                            catch (Exception exception)
                            {
                                HandleException(exception);
                            }
                            finally
                            {
                                if (context != null)
                                {
                                    context.Undo();
                                }
                            }
                        }
                        catch (Exception ex)
                        {
                            this._outputConsole += ex.Message + "<br/>";
                        }
                    }
                }
            }
        }
    }

    #region InfoPath encode and decode classes
    /// <summary>
    /// Decodes a file attachment and saves it to a specified path.
    /// </summary>
    public class InfoPathAttachmentDecoder
    {
        private const int SP1Header_Size = 20;
        private const int FIXED_HEADER = 16;

        private int fileSize;
        private int attachmentNameLength;
        private string attachmentName;
        private byte[] decodedAttachment;

        /// <summary>
        /// Accepts the Base64 encoded string
        /// that is the attachment.
        /// </summary>
        /// <param name="theBase64EncodedString">string to decode</param>
        public InfoPathAttachmentDecoder(string theBase64EncodedString)
        {
            byte[] theData = Convert.FromBase64String(theBase64EncodedString);
            using (MemoryStream ms = new MemoryStream(theData))
            {
                BinaryReader theReader = new BinaryReader(ms);
                DecodeAttachment(theReader);
            }
        }

        /// <summary>
        /// decode the attachment
        /// </summary>
        /// <param name="theReader">a binary reader object</param>
        private void DecodeAttachment(BinaryReader theReader)
        {
            //Position the reader to get the file size.
            byte[] headerData = new byte[FIXED_HEADER];
            headerData = theReader.ReadBytes(headerData.Length);

            fileSize = (int)theReader.ReadUInt32();
            attachmentNameLength = (int)theReader.ReadUInt32() * 2;

            byte[] fileNameBytes = theReader.ReadBytes(attachmentNameLength);

            //InfoPath uses UTF8 encoding.
            Encoding enc = Encoding.Unicode;
            attachmentName = enc.GetString(fileNameBytes, 0, attachmentNameLength - 2);
            decodedAttachment = theReader.ReadBytes(fileSize);
        }

        // commented out because filesystem is disabled for this process
        /*public void SaveAttachment(string saveLocation)
        {
            string fullFileName = saveLocation;
            if (!fullFileName.EndsWith(Path.DirectorySeparatorChar.ToString()))
            {
                fullFileName += Path.DirectorySeparatorChar;
            }

            fullFileName += attachmentName;

            if (File.Exists(fullFileName))
                File.Delete(fullFileName);

            FileStream fs = new FileStream(fullFileName, FileMode.CreateNew);
            BinaryWriter bw = new BinaryWriter(fs);
            bw.Write(decodedAttachment);

            bw.Close();
            fs.Close();
        }*/

        /// <summary>
        /// gets Filename
        /// </summary>
        public string Filename
        {
            get { return attachmentName; }
        }

        /// <summary>
        /// gets FileSize
        /// </summary>
        public int FileSize
        {
            get { return fileSize; }
        }

        /// <summary>
        /// gets DecodedAttachement
        /// </summary>
        public byte[] DecodedAttachment
        {
            get { return decodedAttachment; }
        }
    }

    /// <summary>
    /// InfoPathAttachment encodes file data into the format expected by InfoPath for use in file attachment nodes.
    /// </summary>
    public class InfoPathAttachmentEncoder
    {
        private string base64EncodedFile = string.Empty;
        private string fullyQualifiedFileName = string.Empty;
        private byte[] unencodedBase64Bytes = null;
        private string requestedOutputFilename = string.Empty;
        private long originalFileLength = 0;

        /// <summary>
        /// Creates an encoder to create an InfoPath attachment string.
        /// [this constructor needs access to filesystem to work]
        /// </summary>
        /// <param name="fullyQualifiedFileName">the filename of the file to encode</param>
        public InfoPathAttachmentEncoder(string fullyQualifiedFileName)
        {
            if (fullyQualifiedFileName == string.Empty)
                throw new ArgumentException("Must specify file name", "fullyQualifiedFileName");

            if (!File.Exists(fullyQualifiedFileName))
                throw new FileNotFoundException("File does not exist: " + fullyQualifiedFileName, fullyQualifiedFileName);

            this.fullyQualifiedFileName = fullyQualifiedFileName;
        }

        /// <summary>
        /// Creates an encoder to create an InfoPath attachment string from byte array.
        /// </summary>
        /// <param name="unencodedBase64Bytes">the bytes of the file to encode</param>
        /// <param name="requestedOutputFilename">the output filename</param>
        /// <param name="originalFileLength">the length of the file to encode</param>
        public InfoPathAttachmentEncoder(byte[] unencodedBase64Bytes, string requestedOutputFilename, long originalFileLength)
        {
            if (unencodedBase64Bytes == null)
                throw new ArgumentException("Must supply unencoded byte array", "unencodedBase64Bytes");

            if (string.IsNullOrEmpty(requestedOutputFilename))
                throw new ArgumentException("Must supply output filename", "requestedOutputFilename");

            if (string.IsNullOrEmpty(requestedOutputFilename))
                throw new ArgumentException("Must supply file length", "originalFileLength");

            this.unencodedBase64Bytes = unencodedBase64Bytes;
            this.requestedOutputFilename = requestedOutputFilename;
            this.originalFileLength = originalFileLength;
        }

        /// <summary>
        /// Returns a Base64 encoded string.
        /// </summary>
        /// <returns>String</returns>
        public string ToBase64String()
        {
            if (base64EncodedFile != string.Empty)
                return base64EncodedFile;

            #region default file system encoding <as supplied by Microsoft>
            if (fullyQualifiedFileName != string.Empty)
            {
                // This memory stream will hold the InfoPath file attachment buffer before Base64 encoding.
                MemoryStream ms = new MemoryStream();

                // Get the file information.
                using (BinaryReader br = new BinaryReader(File.Open(fullyQualifiedFileName, FileMode.Open, FileAccess.Read, FileShare.Read)))
                {
                    string fileName = Path.GetFileName(fullyQualifiedFileName);

                    uint fileNameLength = (uint)fileName.Length + 1;

                    byte[] fileNameBytes = Encoding.Unicode.GetBytes(fileName);

                    using (BinaryWriter bw = new BinaryWriter(ms))
                    {
                        // Write the InfoPath attachment signature. 
                        bw.Write(new byte[] { 0xC7, 0x49, 0x46, 0x41 });

                        // Write the default header information.
                        bw.Write((uint)0x14);	// size
                        bw.Write((uint)0x01);	// version
                        bw.Write((uint)0x00);	// reserved

                        // Write the file size.
                        bw.Write((uint)br.BaseStream.Length);

                        // Write the size of the file name.
                        bw.Write((uint)fileNameLength);

                        // Write the file name (Unicode encoded).
                        bw.Write(fileNameBytes);

                        // Write the file name terminator. This is two nulls in Unicode.
                        bw.Write(new byte[] { 0, 0 });

                        // Iterate through the file reading data and writing it to the outbuffer.
                        byte[] data = new byte[64 * 1024];
                        int bytesRead = 1;

                        while (bytesRead > 0)
                        {
                            bytesRead = br.Read(data, 0, data.Length);
                            bw.Write(data, 0, bytesRead);
                        }
                    }
                }

                // This memorystream will hold the Base64 encoded InfoPath attachment.
                MemoryStream msOut = new MemoryStream();

                using (BinaryReader br = new BinaryReader(new MemoryStream(ms.ToArray())))
                {
                    // Create a Base64 transform to do the encoding.
                    ToBase64Transform tf = new ToBase64Transform();

                    byte[] data = new byte[tf.InputBlockSize];
                    byte[] outData = new byte[tf.OutputBlockSize];

                    int bytesRead = 1;

                    while (bytesRead > 0)
                    {
                        bytesRead = br.Read(data, 0, data.Length);

                        if (bytesRead == data.Length)
                            tf.TransformBlock(data, 0, bytesRead, outData, 0);
                        else
                            outData = tf.TransformFinalBlock(data, 0, bytesRead);

                        msOut.Write(outData, 0, outData.Length);
                    }
                }

                msOut.Close();

                return base64EncodedFile = Encoding.ASCII.GetString(msOut.ToArray());
            }
            #endregion

            #region in-memory encoding
            else if (this.unencodedBase64Bytes != null)
            {
                // This memory stream will hold the InfoPath file attachment buffer before Base64 encoding.
                MemoryStream ms = new MemoryStream();

                using (BinaryWriter bw = new BinaryWriter(ms))
                {
                    // Write the InfoPath attachment signature. 
                    bw.Write(new byte[] { 0xC7, 0x49, 0x46, 0x41 });

                    // Write the default header information.
                    bw.Write((uint)0x14);	// size
                    bw.Write((uint)0x01);	// version
                    bw.Write((uint)0x00);	// reserved

                    // Write the file size.
                    bw.Write((uint)this.originalFileLength);

                    uint fileNameLength = (uint)this.requestedOutputFilename.Length + 1;

                    // Write the size of the file name.
                    bw.Write((uint)fileNameLength);

                    // encode the filename (Unicode)
                    byte[] fileNameBytes = Encoding.Unicode.GetBytes(this.requestedOutputFilename);

                    // Write the file name (Unicode encoded).
                    bw.Write(fileNameBytes);

                    // Write the file name terminator. This is two nulls in Unicode.
                    bw.Write(new byte[] { 0, 0 });

                    // write the bytes (supplied)
                    bw.Write(this.unencodedBase64Bytes);
                }

                // This memorystream will hold the Base64 encoded InfoPath attachment.
                MemoryStream msOut = new MemoryStream();

                using (BinaryReader br = new BinaryReader(new MemoryStream(ms.ToArray())))
                {
                    // Create a Base64 transform to do the encoding.
                    ToBase64Transform tf = new ToBase64Transform();

                    byte[] data = new byte[tf.InputBlockSize];
                    byte[] outData = new byte[tf.OutputBlockSize];

                    int bytesRead = 1;

                    while (bytesRead > 0)
                    {
                        bytesRead = br.Read(data, 0, data.Length);

                        if (bytesRead == data.Length)
                            tf.TransformBlock(data, 0, bytesRead, outData, 0);
                        else
                            outData = tf.TransformFinalBlock(data, 0, bytesRead);

                        msOut.Write(outData, 0, outData.Length);
                    }
                }

                msOut.Close();

                return base64EncodedFile = Encoding.ASCII.GetString(msOut.ToArray());
            }
            #endregion
            else
            {
                // something has gong wrong, so return empty string
                return string.Empty;
            }
        }
    }
    #endregion
}
